/*************************************************************************

  riff.c v1.0

    To manage RIFF files.
    riff.c is part of the Oankali GP32 run-time library.

  Copyright (c) 2005 by
  Guillem Reinoso <oankali@nekanium.com>
  www.nekanium.com

  These software is freeware and is provided as is.
  If it doesn't work for you or it makes your PC or GP32 melt down, it's not my
  responsibility.
  If you use it in your programs, cite me please.

**************************************************************************/


#include <gpstdio.h>

#include "riff.h"


RIFF_ErrCode RiffClose (F_HANDLE hFile)
{
  if (GpFileClose(hFile) == SM_OK)
    return riffOK;
  else
    return riffError;
}

RIFF_ErrCode RiffCreate (char *szFilename, FOURCC fccFormType, F_HANDLE *hFile)
{
  F_HANDLE handle;
  DWORD size = sizeof(FOURCC);

  // Create file
  if (GpFileCreate(szFilename, ALWAYS_CREATE, &handle) != SM_OK)
    return riffWriteError;

  // Prepare header
  if (GpFileWrite(handle, "RIFF", sizeof(FOURCC)) != SM_OK) {
    RiffClose(handle);
    return riffWriteError;
  }
  if (GpFileWrite(handle, &size, sizeof(size)) != SM_OK) {
    RiffClose(handle);
    return riffWriteError;
  }
  if (GpFileWrite(handle, &fccFormType, sizeof(fccFormType)) != SM_OK) {
    RiffClose(handle);
    return riffWriteError;
  }

  *hFile = handle;
  return riffOK;
}

RIFF_ErrCode RiffGetChunkData (F_HANDLE hFile, FOURCC fccChunkId,
                               void *vBuffer, DWORD lBufferSize)
{
  RIFF_ErrCode ec;
  DWORD size;
  ulong bytesRead;

  // Get chunk length
  ec = RiffGetChunkSize(hFile, fccChunkId, &size);
  if (ec != riffOK)
    return ec;

  // Get length to get
  if ((lBufferSize != 0) && (lBufferSize < size))
    size = lBufferSize;

  // Read chunk, file position is correct
  GpFileRead(hFile, vBuffer, size, &bytesRead);
  if (bytesRead != (ulong) size)
    return riffCorruptedRIFFFile;

  return riffOK;
}

RIFF_ErrCode RiffGetChunkSize (F_HANDLE hFile, FOURCC fccChunkId, DWORD *lSize)
{
  FOURCC id;
  ulong oldPos;
  ulong bytesRead;
  DWORD size;

  // Go to the beginning of the first chunk
  if (GpFileSeek(hFile, FROM_BEGIN, sizeof(FOURCC) * 2 + sizeof(DWORD), &oldPos) != SM_OK)
    return riffCorruptedRIFFFile;

  // Look for the indicated chunk
  while (TRUE)
  {
    GpFileRead(hFile, &id, sizeof(id), &bytesRead);
    if (bytesRead == 0)
      return riffChunkNotFound;
    if (bytesRead != sizeof(id))
      return riffCorruptedRIFFFile;

    // Reads chunk length
    GpFileRead(hFile, &size, sizeof(size), &bytesRead);
    if (bytesRead != sizeof(size))
      return riffCorruptedRIFFFile;

    // If it's the searched chunk, return length
    if (id == fccChunkId) {
      *lSize = size;
      return riffOK;
    }

    // If it's not, jump to next chunk
    if (GpFileSeek(hFile, FROM_CURRENT, size, &oldPos) != SM_OK)
      return riffCorruptedRIFFFile;
  }
}

RIFF_ErrCode RiffGetFileChunkData (char *szFilename, FOURCC fccFormType, FOURCC fccChunkId,
                                   void *vBuffer, DWORD lBufferSize)
{
  F_HANDLE hFile;
  RIFF_ErrCode ec;
  FOURCC id;

  // Open RIFF file
  if ((ec = RiffOpen(szFilename, &hFile)) != riffOK)
    return ec;

  // Verify file type
  if ((ec = RiffGetFormType(hFile, &id)) != riffOK) {
    RiffClose(hFile);
    return ec;
  }
  if (id != fccFormType) {
    RiffClose(hFile);
    return riffInvalidFormType;
  }

  // Read chunk
  if ((ec = RiffGetChunkData(hFile, fccChunkId, vBuffer, lBufferSize)) != riffOK) {
    RiffClose(hFile);
    return ec;
  }

  // Close RIFF file
  return RiffClose(hFile);
}

RIFF_ErrCode RiffGetFormType (F_HANDLE hFile, FOURCC *fccFormType)
{
  ulong oldPos;
  ulong bytesRead;

  // Set file position
  if (GpFileSeek(hFile, FROM_BEGIN, sizeof(FOURCC) + sizeof(DWORD), &oldPos) != SM_OK)
    return riffCorruptedRIFFFile;

  // Read file form type
  GpFileRead(hFile, fccFormType, sizeof(*fccFormType), &bytesRead);
  if (bytesRead != sizeof(*fccFormType))
    return riffCorruptedRIFFFile;

  return riffOK;
}

RIFF_ErrCode RiffOpen (char *szFilename, F_HANDLE *hFile)
{
  F_HANDLE handle;
  FOURCC id;
  DWORD size;
  ulong bytesRead;
  ulong posBegin, posEnd;

  // Open file
  if (GpFileOpen(szFilename, OPEN_R | OPEN_W, &handle) != SM_OK)
    return riffFileNotFound;

  // Verify file signature, has to be "RIFF"
  GpFileRead(handle, &id, sizeof(id), &bytesRead);
  if (bytesRead != sizeof(id)) {
    RiffClose(handle);
    return riffNotRIFFFile;
  }
  if (id != MAKEFOURCC('R','I','F','F')) {
    RiffClose(handle);
    return riffNotRIFFFile;
  }

  // Verify data length
  //- read indicated length
  GpFileRead(handle, &size, sizeof(size), &bytesRead);
  if (bytesRead != sizeof(size)) {
    RiffClose(handle);
    return riffCorruptedRIFFFile;
  }

  //- get file length
  GpFileSeek(handle, FROM_END, 0L, &posBegin);
  GpFileSeek(handle, FROM_BEGIN, posBegin, &posEnd);

  //- compare lengths
  if (posEnd - posBegin != size) {
    RiffClose(handle);
    return riffCorruptedRIFFFile;
  }

  *hFile = handle;
  return riffOK;
}

RIFF_ErrCode RiffSaveChunkData (F_HANDLE hFile, FOURCC fccChunkId,
                                void *vBuffer, DWORD lBufferSize)
{
  RIFF_ErrCode ec;
  DWORD size;
  ulong oldPos;

  // Look for the indicated chunk
  ec = RiffGetChunkSize(hFile, fccChunkId, &size);
  if ((ec != riffOK) && (ec != riffChunkNotFound)) {
    RiffClose(hFile);
    return ec;
  }

  if (ec == riffOK) {
    // If found chunk is to small, cancel it
    if (size < lBufferSize) {
      // Go back
      if (GpFileSeek(hFile, FROM_BEGIN, 0L, &oldPos) != SM_OK)
        return riffReadError;
      if (GpFileSeek(hFile, FROM_BEGIN, oldPos - (sizeof(FOURCC) + sizeof(DWORD)), &oldPos) != SM_OK)
        return riffReadError;

      // Cancel chunk
      if (GpFileWrite(hFile, "????", sizeof(FOURCC)) != SM_OK)
        return riffWriteError;

      // Go to end of file
      if (GpFileSeek(hFile, FROM_END, 0L, &oldPos) != SM_OK)
        return riffReadError;
    }
    // If the correct chunk is found, save it, file position is correct
    else
      if (GpFileWrite(hFile, vBuffer, lBufferSize) != SM_OK)
        return riffWriteError;
      else
        return riffOK;
  }

  // Creates a new chunk, file position is correct
  if (GpFileWrite(hFile, &fccChunkId, sizeof(fccChunkId)) != SM_OK)
    return riffWriteError;
  if (GpFileWrite(hFile, &lBufferSize, sizeof(lBufferSize)) != SM_OK)
    return riffWriteError;
  if (GpFileWrite(hFile, vBuffer, lBufferSize) != SM_OK)
    return riffWriteError;

  // Update file header
  //- bak to the beginning
  if (GpFileSeek(hFile, FROM_BEGIN, sizeof(FOURCC), &oldPos) != SM_OK)
    return riffReadError;
  size = oldPos - (sizeof(FOURCC) + sizeof(DWORD));
  if (GpFileWrite(hFile, &size, sizeof(size)) != SM_OK)
    return riffWriteError;

  return riffOK;
}

RIFF_ErrCode RiffSaveFileChunkData (char *szFilename, FOURCC fccFormType, FOURCC fccChunkId,
                                    void *vBuffer, DWORD lBufferSize)
{
  F_HANDLE hFile;
  RIFF_ErrCode ec;
  FOURCC id;

  // Try to open RIFF file
  ec = RiffOpen(szFilename, &hFile);
  if ((ec != riffOK) && (ec != riffFileNotFound))
    return ec;

  // If file not found, create it
  if (ec == riffFileNotFound) {
    if ((ec = RiffCreate(szFilename, fccFormType, &hFile)) != riffOK)
      return ec;
  }
  // If found, verify it
  else {
    // Verify file form type
    if ((ec = RiffGetFormType(hFile, &id)) != riffOK) {
      RiffClose(hFile);
      return ec;
    }
    if (id != fccFormType) {
      RiffClose(hFile);
      return riffInvalidFormType;
    }
  }

  // Save chunk
  if ((ec = RiffSaveChunkData(hFile, fccChunkId, vBuffer, lBufferSize)) != riffOK) {
    RiffClose(hFile);
    return ec;
  }

  // Close RIFF file
  return RiffClose(hFile);
}
